/**CHeaderFile*****************************************************************

  FileName    [cudd.h]

  PackageName [cudd]

  Synopsis    [The University of Colorado decision diagram package.]

  Description [External functions and data strucures of the CUDD package.
  <ul>
  <li> To turn on the gathering of statistics, define DD_STATS.
  <li> To link with mis, define DD_MIS.
  </ul>
  Modified by Abelardo Pardo to interface it to VIS.
  ]

  SeeAlso     []

  Author      [Fabio Somenzi]

  Copyright [This file was created at the University of Colorado at
  Boulder.  The University of Colorado at Boulder makes no warranty
  about the suitability of this software for any purpose.  It is
  presented on an AS IS basis.]

  Revision    [$Id: cudd.h,v 1.89 1997/04/13 17:29:12 fabio Exp $]

  Revision 2 Oct 23rd, Yvonne Coady, changing into DdNodePar, DdNodeOld
             Motivation:  in order to parallelize the BDD package, we
             had to introduce the notion of "level" -- besides, Kim did
             not like the pointer to the DdNode just hanging around anyway...

             Changes include:
                      struct DdNodePar
                      constant CUDD_ERROR
                      DdNodePar for kids (T and E)
                      DdNodeOld (replacing DdNode)
                      macro Cudd_IsConstantPar
                      function Cudd_NotPar returns DdNodePar
                      macro Cudd_NotCondPar
                      macro Cudd_RegularPar
                      function Cudd_ComplementPar returns DdNodePar
                      macro Cudd_IsComplementPar
                      macro Cudd_T_Par
                      macro Cudd_E_Par
                      various EXTERN declarations...

  Revision 3 start jan 20
             changing memory management

******************************************************************************/

#ifndef _CUDD
#define _CUDD

/*---------------------------------------------------------------------------*/
/* Nested includes                                                           */
/*---------------------------------------------------------------------------*/

/* #include "mtr.h"  don't need to be grouping anything... */
#include <stdio.h>
/* For Kim, who was quite peeved that Fabio did not do this originally */

#include "util.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

#define CUDD_VERSION "2.1.2"



#ifndef SIZEOF_VOID_P
#define SIZEOF_VOID_P 4
#endif
#ifndef SIZEOF_INT
#define SIZEOF_INT 4
#endif
#ifndef SIZEOF_LONG
#define SIZEOF_LONG 4
#endif

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#define CUDD_VALUE_TYPE		double
#define CUDD_OUT_OF_MEM		-1
/* The initial size of the subtables is chosen as the largest prime
** that is less than a given power of two. Valid choices are 61, 127,
** 251, 509, 1021, ... This is for memory efficiency and good hash
** table utilization.
*/
#define CUDD_UNIQUE_SLOTS	251	/* initial size of subtables */
#define CUDD_CACHE_SLOTS	16381	/* default size of the cache */
#define CUDD_MAX_CACHE_SIZE	100000	/* do not resize cache if above
					   this threshold (default) */

/* Reordering types. */
#define CUDD_REORDER_SAME		0
#define CUDD_REORDER_NONE		1
#define CUDD_REORDER_RANDOM		2
#define CUDD_REORDER_RANDOM_PIVOT	3
#define CUDD_REORDER_SIFT		4
#define CUDD_REORDER_SIFT_CONVERGE	5
#define CUDD_REORDER_SYMM_SIFT		6
#define CUDD_REORDER_SYMM_SIFT_CONV	7
#define	CUDD_REORDER_WINDOW2		8
#define	CUDD_REORDER_WINDOW3		9
#define	CUDD_REORDER_WINDOW4		10
#define CUDD_REORDER_WINDOW2_CONV	11
#define CUDD_REORDER_WINDOW3_CONV	12
#define CUDD_REORDER_WINDOW4_CONV	13
#define CUDD_REORDER_GROUP_SIFT		14
#define CUDD_REORDER_GROUP_SIFT_CONV	15
#define CUDD_REORDER_ANNEALING		16
#define CUDD_REORDER_GENETIC		17
#define CUDD_REORDER_LINEAR		18
#define CUDD_REORDER_LINEAR_CONVERGE	19
#define CUDD_REORDER_EXACT		20

/* Aggregation methods. */
#define CUDD_NO_CHECK     0
#define CUDD_GROUP_CHECK  1
#define CUDD_GROUP_CHECK2 2
#define CUDD_GROUP_CHECK3 3
#define CUDD_GROUP_CHECK4 4
#define CUDD_GROUP_CHECK5 5
#define CUDD_GROUP_CHECK6 6
#define CUDD_GROUP_CHECK7 7
#define CUDD_GROUP_CHECK8 8
#define CUDD_GROUP_CHECK9 9

/* Constants for residue functions. */
#define CUDD_RESIDUE_DEFAULT	0
#define CUDD_RESIDUE_MSB	1
#define CUDD_RESIDUE_TC		2

/* Hooks. */
#define CUDD_PRE_GC_HOOK		0
#define CUDD_POST_GC_HOOK		1
#define CUDD_POST_REORDERING_HOOK	2

/* CUDD_MAXINDEX is defined in such a way that on 32-bit and 64-bit
** machines one can cast an index to (int) without generating a negative
** number.
** Kim: only works if sizeof(short)<sizeof(int)
*/
#if SIZEOF_VOID_P == 8 && SIZEOF_INT == 4
#define CUDD_MAXINDEX             (((DdHalfWord) ~0) >> 1)
#else
#define CUDD_MAXINDEX             ((DdHalfWord) ~0)
#endif

/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

#if SIZEOF_VOID_P == 8 && SIZEOF_INT == 4
typedef unsigned int   DdHalfWord;
typedef int DdSignedHalfWord;
#else
typedef unsigned short DdHalfWord;
typedef short DdSignedHalfWord;
#endif

#ifdef __osf__
#pragma pointer_size save
#pragma pointer_size short
#endif

/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  Rev2:DdNodePar has been introduced to this parallel version, and the 
        previous version of a DdNode is now "DdNodeOld"
  Rev3:pointers are now converted to machine independant adresses

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

typedef struct DdNodeAdr
{
  DdHalfWord page;
  DdSignedHalfWord number;
} DdNodeAdr;

typedef struct DdNodePar {
    DdHalfWord index;
    DdNodeAdr  address;
} DdNodePar;

typedef struct DdNode {
    DdHalfWord   index;
    DdNodePar    NodePar;
    DdHalfWord   ref;
    struct DdNode *next;
} DdNode;

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ 
  Rev2: DdNodePar is now used for the children as well    
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
  
typedef struct DdChildren {
    DdNodePar T;
    DdNodePar E;
} DdChildren;


/* The DdNode structure is the only one exported out of the package */

/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ 
   Rev2: DdNode is now DdNodeOld, and is no longer exported...
         next is now a pointer to a DdNodeOld    
   Rev3: refrence count removed
         index removed
         Next is now machine independant adress, because getting the
         machine independant address from the real adress is hard
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

typedef struct DdNodeOld {
    DdNodeAdr next;     	/* next pointer for unique table */
    union {
	CUDD_VALUE_TYPE value;	/* for constant nodes */
	DdChildren kids;	/* for internal nodes */
    } type;
} DdNodeOld;

typedef union DdNodeMore {
  struct {
    DdHalfWord Ref;
    DdHalfWord Index;
    DdHalfWord NextIndex;
  } Reorder;
    DdNodePar NewLocation;
} DdNodeMore;



#ifdef __osf__
#pragma pointer_size restore
#endif

typedef struct DdManager DdManager;

typedef struct DdGen DdGen;

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

#ifdef DD_STATS
#ifdef CUDD_MAIN
int cudd_Nodesfreed;		/* number of nodes returned to the free list */
int cudd_Nodesdropped;		/* number of nodes killed by dereferencing */
#else
extern int cudd_Nodesfreed;
extern int cudd_Nodesdropped;
#endif
#endif

#ifdef DD_UNIQUE_PROFILE
#ifdef CUDD_MAIN
int cudd_UniqueLookUps;		/* number of unique table lookups */
int cudd_UniqueLinks;		/* total distance traveled in coll. chains */
#else
extern int cudd_UniqueLookUps, cudd_UniqueLinks;
#endif
#endif

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ 
   Rev2:  Most of these macros have been redone as functions, dealing
   with DdNodePar structures instead of DdNode pointers.
   NOTE: Original versions of these macros have been commented out!
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/**Macro***********************************************************************

  Synopsis     [Returns 1 if the node is a constant node.]

  Description  [Returns 1 if the node is a constant node (rather than an
  internal node). All constant nodes have the same index
  (CUDD_MAXINDEX). The pointer passed to Cudd_IsConstant may be either
  regular or complemented.]

  SideEffects  [none]

  SeeAlso      []

******************************************************************************/

#define Cudd_IsConstant(node) ((Cudd_Regular(node))->index == CUDD_MAXINDEX)

/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ 
   Rev2: New marcro for Par nodes  -- level is used instead of index  
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

#define Cudd_IsConstantPar(nodePar) ( nodePar.level == CUDD_MAXINDEX )


/**Macro***********************************************************************

  Synopsis     [Complements a DD.]

  Description  [Complements a DD by flipping the complement attribute of
  the pointer (the least significant bit).]

  SideEffects  [none]

  SeeAlso      [Cudd_NotCond]

******************************************************************************/

#define Cudd_Not(node) ((DdNode *)((long)(node) ^ 01))

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ 
   Rev2 New function for Par nodes 
   Rev3 reintroducing old macros
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

DdNodePar Cudd_NotPar( DdNodePar nodePar );


/**Macro***********************************************************************

  Synopsis     [Complements a DD if a condition is true.]

  Description  [Complements a DD if condition c is true; c should be
  either 0 or 1, because it is used directly (for efficiency). If in
  doubt on the values c may take, use "(c) ? Cudd_Not(node) : node".]

  SideEffects  [none]

  SeeAlso      [Cudd_Not]

******************************************************************************/

#define Cudd_NotCond(node,c) ((DdNode *)((long)(node) ^ (c)))


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   Rev2 New macro for Par nodes 
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

#define Cudd_NotCondPar( nodePar, c ) ( (c)? (Cudd_NotPar( nodePar )) : (nodePar) )


/**Macro***********************************************************************

  Synopsis     [Returns the regular version of a pointer.]

  Description  []

  SideEffects  [none]

  SeeAlso      [Cudd_Complement Cudd_IsComplement]

******************************************************************************/

#define Cudd_Regular(node) ((DdNode *)((unsigned long)(node) & ~01))



/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  Rev2 New function for Par nodes 
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

#define Cudd_RegularPar( nodePar ) ((DdNodeOld *)((unsigned long) (nodePar.node) & ~01) NOTE WRONG)


/**Macro***********************************************************************

  Synopsis     [Returns the complemented version of a pointer.]

  Description  []

  SideEffects  [none]

  SeeAlso      [Cudd_Regular Cudd_IsComplement]

******************************************************************************/

#define Cudd_Complement(node) ((DdNode *)((unsigned long)(node) | 01))


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   Rev2 New function for Par nodes 
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

DdNodePar Cudd_ComplementPar( DdNodePar nodePar );

/**Macro***********************************************************************

  Synopsis     [Returns 1 if a pointer is complemented.]

  Description  []

  SideEffects  [none]

  SeeAlso      [Cudd_Regular Cudd_Complement]

******************************************************************************/

#define Cudd_IsComplement(node)	((long) (node) & 01)



/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   Rev2 New macro for Par nodes 
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

#define Cudd_IsComplementPar(nodePar) ((nodePar).address.number < 0)


/**Macro***********************************************************************

  Synopsis     [Returns the then child of an internal node.]

  Description  [Returns the then child of an internal node. If
  <code>node</code> is a constant node, the result is unpredictable.]

  SideEffects  [none]

  SeeAlso      [Cudd_E Cudd_V]

******************************************************************************/
/*
#define Cudd_T(node) ((Cudd_Regular(node))->type.kids.T)
*/


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   Rev2 New macro for Par nodes, returns a DdNodePar (the then child) 
   Rev3 New adress types
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

#define Cudd_T_Par(unique,node) ((unique)->memory[(node).index].PageTable[(node).address.page][abs((node).address.number)].type.kids.T)


/**Macro***********************************************************************

  Synopsis     [Returns the else child of an internal node.]

  Description  [Returns the else child of an internal node. If
  <code>node</code> is a constant node, the result is unpredictable.]

  SideEffects  [none]

  SeeAlso      [Cudd_T Cudd_V]

******************************************************************************/
/*
#define Cudd_E(node) ((Cudd_Regular(node))->type.kids.E)
*/


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   Rev2 New macro for Par nodes 
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

#define Cudd_E_Par(unique,node) ((unique)->memory[(node).index].PageTable[(node).address.page][abs((node).address.number)].type.kids.E)

#define Cudd_V_Par(unique,node) ((unique)->memory[(node).index].PageTable[(node).address.page][abs((node).address.number)].type.value)


/**Macro***********************************************************************

  Synopsis     [Returns the current position in the order of variable
  index.]

  Description [Returns the current position in the order of variable
  index. This macro is obsolete and is kept for compatibility. New
  applications should use Cudd_ReadPerm instead.]

  SideEffects  [none]

  SeeAlso      [Cudd_ReadPerm]

******************************************************************************/
/*
#define Cudd_ReadIndex(dd,index) (Cudd_ReadPerm(dd,index))
*/

/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   Rev 2: What the heck -- I don't know what ReadPerm is anyway???
   This has not yet been done!!!!
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/* These are potential duplicates. */
#ifndef EXTERN
#   ifdef __cplusplus
#	define EXTERN extern "C"
#   else
#	define EXTERN extern
#   endif
#endif
#ifndef ARGS
#   if defined(__STDC__) || defined(__cplusplus)
#	define ARGS(protos)	protos		/* ANSI C */
#   else /* !(__STDC__ || __cplusplus) */
#	define ARGS(protos)	()		/* K&R C */
#   endif /* !(__STDC__ || __cplusplus) */
#endif


/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Function prototypes                                                       */
/*---------------------------------------------------------------------------*/


EXTERN DdNode* Cudd_bddNewVar ARGS((DdManager *dd));

EXTERN DdNode* Cudd_bddIthVar ARGS((DdManager *dd, int i));

EXTERN DdNode* Cudd_ReadOne ARGS((DdManager *dd));

EXTERN DdNode* Cudd_ReadLogicZero ARGS((DdManager *dd));

EXTERN int Cudd_ReadSize ARGS((DdManager *dd));

EXTERN unsigned int Cudd_NodeReadIndex ARGS((DdNode *node));

EXTERN int Cudd_ReadPerm ARGS((DdManager *dd, int i));

EXTERN int Cudd_ReadInvPerm ARGS((DdManager *dd, int i));

EXTERN DdNode* Cudd_ReadVars ARGS((DdManager *dd, int i));

EXTERN int Cudd_ReadGroupcheck ARGS((DdManager *dd));

EXTERN void Cudd_SetGroupcheck ARGS((DdManager *dd, int gc));

EXTERN long Cudd_ReadMemoryInUse ARGS((DdManager *dd));
EXTERN int Cudd_PrintInfo ARGS((DdManager *dd, FILE *fp));
EXTERN long Cudd_ReadPeakNodeCount ARGS((DdManager *dd));
EXTERN long Cudd_ReadNodeCount ARGS((DdManager *dd));

EXTERN DdNodePar Cudd_bddBooleanDiff ARGS((DdManager *manager, DdNodePar f, int x));
EXTERN double Cudd_bddCorrelation ARGS((DdManager *manager, DdNodePar f, DdNodePar g));
EXTERN double Cudd_bddCorrelationWeights ARGS((DdManager *manager, DdNodePar f, DdNodePar g, double *prob));
EXTERN DdNode* Cudd_bddIte ARGS((DdManager *dd, DdNode* f, DdNode* g, DdNode* h));
EXTERN DdNode* Cudd_bddAnd ARGS((DdManager *dd, DdNode* f, DdNode* g));
EXTERN DdNode* Cudd_bddOr ARGS((DdManager *dd, DdNode* f, DdNode* g));
EXTERN DdNode* Cudd_bddNand ARGS((DdManager *dd, DdNode* f, DdNode* g));
EXTERN DdNode* Cudd_bddNor ARGS((DdManager *dd, DdNode* f, DdNode* g));
EXTERN DdNode* Cudd_bddXor ARGS((DdManager *dd, DdNode* f, DdNode* g));
EXTERN DdNode* Cudd_bddXnor ARGS((DdManager *dd, DdNode* f, DdNode* g));

EXTERN int Cudd_DebugCheck ARGS((DdManager *table));
EXTERN int Cudd_CheckKeys ARGS((DdManager *table));
EXTERN DdNodePar Cudd_Cofactor ARGS((DdManager *dd, DdNodePar f, DdNodePar g));

EXTERN DdManager * Cudd_Init ARGS((unsigned int numVars, unsigned int numVarsZ, unsigned int numSlots, unsigned int cacheSize, unsigned int maxCacheSize));

/*  Cache stuff
EXTERN unsigned int Cudd_ReadCacheSlots ARGS((DdManager *dd));
EXTERN unsigned int Cudd_ReadCacheLookUps ARGS((DdManager *dd));
EXTERN unsigned int Cudd_ReadCacheHits ARGS((DdManager *dd));
EXTERN unsigned int Cudd_ReadMaxCache ARGS((DdManager *dd));
EXTERN void Cudd_SetMaxCache ARGS((DdManager *dd, unsigned int mc));
EXTERN unsigned int Cudd_ReadMaxCacheHard ARGS((DdManager *dd));
EXTERN void Cudd_SetMaxCacheHard ARGS((DdManager *dd, unsigned int mc));
*/


EXTERN void Cudd_Quit ARGS((DdManager *unique));

EXTERN int Cudd_bddRead ARGS((FILE *fp, DdManager *dd, DdNodePar *E, DdNodePar **x, DdNodePar **y, int *nx, int *ny, int *m, int *n, int bx, int sx, int by, int sy));
EXTERN void Cudd_Ref ARGS((DdNode *n));
EXTERN void Cudd_RecursiveDeref ARGS((DdManager *table, DdNode *n));

EXTERN void Cudd_Deref ARGS((DdNode *node));
EXTERN int Cudd_CheckZeroRef ARGS((DdManager *manager));
EXTERN int Cudd_ReduceHeap ARGS((DdManager *table, int heuristic, int minsize));
EXTERN int Cudd_ShuffleHeap ARGS((DdManager *table, int *permutation));
EXTERN DdNodePar Cudd_Eval ARGS((DdManager *dd, DdNodePar f, int *inputs));
EXTERN DdNodePar Cudd_ShortestPath ARGS((DdManager *manager, DdNodePar f, int *weight, int *support, int *length));

EXTERN unsigned int Cudd_Prime ARGS((unsigned int p));
EXTERN int Cudd_PrintMinterm ARGS((DdManager *manager, DdNodePar node));
EXTERN int Cudd_PrintDebug ARGS((DdManager *dd, DdNodePar f, int n, int pr));
EXTERN int Cudd_DagSize ARGS((DdNodePar node));
EXTERN int Cudd_SharingSize ARGS((DdNodePar *nodeArray, int n));

/**AutomaticEnd***************************************************************/
/* general use */
#define CUDD_BDD_ITE        1
#define CUDD_BDD_AND        2
#define CUDD_BDD_XOR        3
#define CUDD_BDD_INTERSECT  4
#define CUDD_BDD_EXISTS     5
#define CUDD_BDD_EXISTS_PRE 6
#define CUDD_BDD_ANDABSTRACT 7
#define CUDD_BDD_ANDABSTRACT_PRE 8

/* 100 series reserved for message types */

/* only used in the initial call */
#define CUDD_BDD_OR   201
#define CUDD_BDD_NOR  202
#define CUDD_BDD_NAND 203
#define CUDD_BDD_XNOR 204
#define CUDD_BDD_FORALL 205

/* special use in connection with Ite Base */
#define CUDD_BDD_OR_1_3      301
#define CUDD_BDD_OR_NOT1_2   302
#define CUDD_BDD_AND_1_2     303
#define CUDD_BDD_AND_NOT1_3  304

#define cuddEqualAdr(a,b)  ((a).number==(b).number && (a).page==(b).page)
/* compare number first because they will more often be different */ 

#define cuddEqualPar(f,g)  (cuddEqualAdr((f).address,(g).address) && (f).index==(g).index)

#define cuddEqual(f,g)  ((f)==(g))
#define NULL_USER NULL

#endif /* _CUDD */

