/**CFile***********************************************************************

  FileName    [cuddCache.c]

  PackageName [cudd]

  Synopsis    [Functions for cache insertion and lookup.]

  Description [Internal procedures included in this module:
		<ul>
		<li> cuddInitCache()
		<li> cuddCacheInsert()
		<li> cuddCacheInsert2()
		<li> cuddCacheLookup()
		<li> cuddCacheLookupZdd()
		<li> cuddCacheLookup2()
		<li> cuddCacheLookup2Zdd()
		<li> cuddConstantLookup()
		<li> cuddCacheProfile()
		</ul>
	    Static procedures included in this module:
		<ul>
		<li> ddCacheResize()
		</ul> ]

  SeeAlso     []

  Author      [Fabio Somenzi]

  Copyright [ This file was created at the University of Colorado at
  Boulder.  The University of Colorado at Boulder makes no warranty
  about the suitability of this software for any purpose.  It is
  presented on an AS IS basis.]

******************************************************************************/

#include    "util.h"
#include    "cuddInt.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

#ifndef lint
static char rcsid[] DD_UNUSED = "$Id: cuddCache.c,v 1.12 1997/02/15 22:26:04 fabio Exp $";
#endif

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/


/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

/**AutomaticEnd***************************************************************/


/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/


/**Function********************************************************************

  Synopsis    [Initializes the computed table.]

  Description [Initializes the computed table. It is called by
  Cudd_Init. Returns 1 in case of success; 0 otherwise.]

  SideEffects [None]

  SeeAlso     [Cudd_Init]

******************************************************************************/
int
cuddInitCachePar(unique,cacheSize,maxCacheSize)
DdManager *unique;		/* unique table */
unsigned int cacheSize;		/* Initial size of the cache */
unsigned int maxCacheSize;	/* Size of the cache beyond which */
				/* no resizing occurs */
{
    int i;

    cacheSize = Cudd_Prime(cacheSize);
    unique->cache = ALLOC(DdCache,cacheSize);
    if (unique->cache == NULL) return(0);
    unique->cacheSlots = cacheSize;
    /* Initialize to avoid division by 0 and immediate resizing. */
    unique->cacheLookUps = cacheSize * DD_MIN_HIT / 100 + 1;
    unique->cacheHits = 0;
    unique->minHit = DD_MIN_HIT;
    unique->maxCacheHard = maxCacheSize;
    unique->maxCache = ddMin(maxCacheSize,unique->slots);
    unique->memused += cacheSize * sizeof(DdCache);

    /* Initialize the cache */
    for (i = 0; (unsigned) i < cacheSize; i++) {
	unique->cache[i].f = CUDD_ERROR;
	/* note f=CUDD_ERROR now represents open spots */
#ifdef DD_CACHE_PROFILE
	unique->cache[i].count = 0;
#endif
    }

    return(1);

} /* end of ddInitCache */

#ifdef DD_CACHE_PROFILE

#define DD_HYSTO_BINS 10

/**Function********************************************************************

  Synopsis    [Computes and prints a profile of the cache usage.]

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/
void
cuddCacheProfilePar(table)
DdManager *table;
{
    DdCache *cache;
    double count, mean, meansq, stddev, expected;
    int max, min, imax, imin;
    int i, slots;
    int *hystogram;
    int nbins = DD_HYSTO_BINS;
    int bin;
    int thiscount, totalcount;
    int nzeroes;

    cache = table->cache;
    slots = table->cacheSlots;

    meansq = mean = expected = 0.0;
    max = min = cache[0].count;
    imax = imin = nzeroes = totalcount = 0;

    hystogram = ALLOC(int, nbins);
    for (i = 0; i < nbins; i++) {
	hystogram[i] = 0;
    }

    for (i = 0; i < slots; i++) {
	thiscount = cache[i].count;
	if (thiscount > max) {
	    max = thiscount;
	    imax = i;
	}
	if (thiscount < min) {
	    min = thiscount;
	    imin = i;
	}
	if (thiscount == 0) {
	    nzeroes++;
	}
	count = (double) thiscount;
	mean += count;
	meansq += count * count;
	totalcount += thiscount;
	expected += count * (double) i;
	bin = (i * nbins) / slots;
	hystogram[bin] += thiscount;
    }
    mean /= (double) slots;
    meansq /= (double) slots;
    stddev = sqrt(meansq - mean*mean);

    (void) fprintf(stdout,"# Cache stats: slots = %d average = %g ", slots, mean);
    (void) fprintf(stdout,"standard deviation = %g\n", stddev);
    (void) fprintf(stdout,"# Cache max accesses = %d for slot %d\n", max, imax);
    (void) fprintf(stdout,"# Cache min accesses = %d for slot %d\n", min, imin);
    (void) fprintf(stdout,"# Cache unused slots = %d\n", nzeroes);

    if (totalcount) {
	expected /= (double) totalcount;
	(void) fprintf(stdout,"# Cache access hystogram for %d bins", nbins);
	(void) fprintf(stdout," (expected bin value = %g)\n# ", expected);
	for (i = nbins - 1; i>=0; i--) {
	    (void) fprintf(stdout,"%d ", hystogram[i]);
	}
	(void) fprintf(stdout,"\n");
    }

    FREE(hystogram);

} /* end of cuddCacheProfile */
#endif

/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Resizes the cache.]

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/









