/**CFile***********************************************************************

  FileName    [cuddRef.c]

  PackageName [cudd]

  Synopsis    [Functions that manipulate the reference counts.]

  Description [External procedures included in this module:
		    <ul>
		    <li> Cudd_Ref()
		    <li> Cudd_RecursiveDeref()
		    <li> Cudd_RecursiveDerefZdd()
		    <li> Cudd_Deref()
		    <li> Cudd_CheckZeroRef()
		    </ul>
	       Internal procedures included in this module:
		    <ul>
		    <li> cuddReclaim()
		    <li> cuddReclaimZdd()
		    </ul>
	       Static procedures included in this module:
		    <ul>
		    <li> ddDoReclaim()
		    <li> ddDoReclaimZdd()
		    </ul>
	      ]

  SeeAlso     []

  Author      [Fabio Somenzi]

  Copyright [ This file was created at the University of Colorado at
  Boulder.  The University of Colorado at Boulder makes no warranty
  about the suitability of this software for any purpose.  It is
  presented on an AS IS basis.]

******************************************************************************/

#include    "util.h"
#include    "cuddInt.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

#ifndef lint
static char rcsid[] DD_UNUSED = "$Id: cuddRef.c,v 1.10 1997/04/13 10:16:14 fabio Exp $";
#endif

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/


/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

/**AutomaticEnd***************************************************************/


/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis [Increases the reference count of a node, if it is not
  saturated.]

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/
DD_INLINE

void
Cudd_Ref(DdNode *n)
{
    n = Cudd_Regular(n);
    cuddSatInc(n->ref);
} /* end of Cudd_Ref */


/**Function********************************************************************

  Synopsis    [Decreases the reference count of node n.]

  Description [Decreases the reference count of node n. If n dies,
  recursively decreases the reference counts of its children.  It is
  used to dispose of a DD that is no longer needed.]

  SideEffects [None]

  SeeAlso     [Cudd_Deref]

******************************************************************************/
void
Cudd_RecursiveDeref(table,n)
DdManager *table;
DdNode *n;
{  
    n = Cudd_Regular(n);
    cuddSatDec(n->ref);
}



/**Function********************************************************************

  Synopsis    [Decreases the reference count of node.]

  Description [Decreases the reference count of node. It is primarily
  used in recursive procedures to decrease the ref count of a result
  node before returning it. This accomplishes the goal of removing the
  protection applied by a previous Cudd_Ref.]

  SideEffects [None]

  SeeAlso     [Cudd_RecursiveDeref]

******************************************************************************/
DD_INLINE
void
Cudd_Deref(node)
DdNode *node;
{
    node = Cudd_Regular(node);
    cuddSatDec(node->ref);
} /* end of Cudd_Deref */


int cuddResetRefs(DdManager *manager)
{
  int page_number,i;
  DdNodeMore *ref_page;
  DdMemory *mem;
  int index;
  
  for(index=0;index<manager->size;index++)
    {
      mem = &manager->memory[index];
      
      if(cuddLocal(manager,index))
	{	  
	  for(page_number=0;page_number<mem->nextPage;page_number++)
	    { 
	      ref_page=mem->PageTable2[page_number];
	      for(i=1;i<=DD_MEM_CHUNK;i++)
		ref_page[i].Reorder.Ref=0;
	    }
	}
      else
	{
	  for(page_number=0;page_number<mem->TableSize;page_number++)
	    { 
	      ref_page=mem->PageTable2[page_number];
	      if(ref_page!=NULL)
		{
		  free(ref_page+1); 
		  mem->PageTable2[page_number]=NULL;
		  /* NOTE it may be better to reset counts */
		}
	    }
	}
    }
  return(1); /* NO ERROR HAMDELING */
}

/* note this assumes that the the ingoing indeg to a node corosponds to the
   index in the node */
int cuddFindRef(DdManager *manager)
{
  int index,level,i;
  DdNodeMore **table2;
  DdSubtable *subtable;
  DdNodeOld *f_ptr;
  DdNodePar f;
  DdNodeAdr *last,*nodelist;

  for(index=0;index<manager->size;index++) {
    if(cuddLocal(manager,index)) {
      f.index=index;
      subtable=&manager->subtables[index];
      nodelist=subtable->nodelist;
      table2=manager->memory[index].PageTable2;
      
      for(i=0;i<subtable->slots;i++) {
	last=&nodelist[i];
	f.address=*last;
	while (!cuddEqualAdr(f.address,NULL_ADR)) {
	  f_ptr=cuddPtrPar(manager,f);

	  if(table2[f.address.page][f.address.number].Reorder.Ref>0) {
	    cuddIncreaseRefSafe(manager,f_ptr->type.kids.T);
	    cuddIncreaseRefSafe(manager,f_ptr->type.kids.E);
	    /* go to next element */
	    last=&f_ptr->next;
	  } else {
	    *last=f_ptr->next; /* remove from table */   
	    subtable->keys--;
	    cuddDeallocNodePar(manager,f);
	  }
	  f.address=*last;
	}
      }
    }
  }
  return(1);
}

/**Function********************************************************************

  Synopsis [Checks the unique table for nodes with non-zero reference
  counts.]

  Description [Checks the unique table for nodes with non-zero
  reference counts. It is normally called before Cudd_Quit to make sure
  that there are no memory leaks due to missing Cudd_RecursiveDeref's.
  Takes into account that reference counts may saturate and that the
  basic constants and the projection functions are referenced by the
  manager.  Returns the number of nodes with non-zero reference count.
  (Except for the cases mentioned above.)]

  SideEffects [None]

  SeeAlso     []

******************************************************************************/
int
Cudd_CheckZeroRef(manager)
DdManager *manager;
{
  /* REV2 nuked */
    return (0);

} /* end of Cudd_CheckZeroRef */


/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/


/**Function********************************************************************

  Synopsis    [Brings children of a dead node back.]

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/
void
cuddReclaim(table, n)
DdManager *table;
DdNodePar n;
{
  /* REV2 nuked */


} /* end of cuddReclaim */



/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

