/**CFile***********************************************************************

  FileName    [cuddTable.c]

  PackageName [cudd]

  Synopsis    [Unique table management functions.]

  Description [External procedures included in this module:
		<ul>
		<li> Cudd_Prime()
		</ul>
	Internal procedures included in this module:
		<ul>
		<li> cuddAllocNode()
		<li> cuddInitTable()
		<li> cuddFreeTable()
		<li> cuddGarbageCollect()
		<li> cuddUniqueInter()
		<li> cuddUniqueConst()
		<li> cuddInsertSubtables()
		<li> cuddDestroySubtables()
		</ul>
	Static procedures included in this module:
		<ul>
		<li> ddRehash()
		</ul>]

  SeeAlso     []

  Author      [Fabio Somenzi]

  Copyright   [This file was created at the University of Colorado at
  Boulder.  The University of Colorado at Boulder makes no warranty
  about the suitability of this software for any purpose.  It is
  presented on an AS IS basis.]

  Revision 2  Sometime in mid November, changes made for distributed CUDD 
              by Kim and Yvonne include:
                    cuddAllocNodePar returns DdNodeOld*
                    cuddInitTablePar
                    cuddFreeTablePar
                    cuddGarbageCollect returns 1
                    cudd UniqueInterPar returns DdNodepar
                    ddResizeTablePar

******************************************************************************/

#include    "util.h"
#include    "cuddInt.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/* REV2 red black tree gone */

/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/

/* This is a hack for when CUDD_VALUE_TYPE is double */
typedef union hack {
    CUDD_VALUE_TYPE value;
    unsigned int bits[2];
} hack;

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

#ifndef lint
static char rcsid[] DD_UNUSED = "$Id: cuddTable.c,v 1.60 1997/04/13 10:16:08 fabio Exp $";
#endif

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/* REV2 red black tree gone */

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static void cuddRemoveFromTable(DdManager *manager,DdNodePar f);
static void ddRehash ARGS((DdManager *unique, int i));
/**AutomaticEnd***************************************************************/


/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/


/**Function********************************************************************

  Synopsis    [Returns the next prime &gt;= p.]

  Description []

  SideEffects [None]

******************************************************************************/
unsigned int
Cudd_Prime(p)
unsigned int p;
{
    int i,pn;

    p--;
    do {
        p++;
        if (p&1) {
	    pn = 1;
	    i = 3;
	    while ((unsigned) (i * i) <= p) {
		if (p % i == 0) {
		    pn = 0;
		    break;
		}
		i += 2;
	    }
	} else {
	    pn = 0;
	}
    } while (!pn);
    return(p);

} /* end of Cudd_Prime */


/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/


int cuddResizePageTable(DdManager *manager, int index)
{
  DdMemory *mem;
  int i,oldSize;
  DdNodeOld **oldTable;
  mem=&manager->memory[index];

  oldSize=mem->TableSize;
  oldTable=mem->PageTable;
  mem->TableSize*=DD_PAGE_GROWTH; 
  mem->PageTable=ALLOC(DdNodeOld *,mem->TableSize);
  if(mem->PageTable == NULL) {
    mem->PageTable=oldTable;
    mem->TableSize=oldSize;
    return(0);
  }

  assert(mem->nextPage == oldSize);
  assert(cuddEqualAdr(mem->nextFree,NULL_ADR));

  for(i=0;i<oldSize;i++)
    mem->PageTable[i]=oldTable[i];

  manager->memused += sizeof(DdNodeOld *) * (mem->TableSize-oldSize);

  if(manager->reordering)
    {
      DdNodeMore **oldTable2;
      oldTable2=mem->PageTable2;
      mem->PageTable2=ALLOC(DdNodeMore *,mem->TableSize);
      if(mem->PageTable2 == NULL) {
	mem->PageTable2=oldTable2;
	mem->TableSize=oldSize;
	return(0);
      }

      for(i=0;i<oldSize;i++)
	mem->PageTable2[i]=oldTable2[i];
    }      

  return 1;
}

/**Function********************************************************************

  Synopsis    [Fast storage allocation for DdNodes in the table.]

  Description [Fast storage allocation for DdNodes in the table. The
  first 4 bytes of a chunk contain a pointer to the next block; the
  rest contains DD_MEM_CHUNK spaces for DdNodes.  Returns a pointer to
  a new node if successful; NULL is memory is full.]

  SideEffects [None]

  SeeAlso     [cuddDynamicAllocNode]

******************************************************************************/
DdNodeAdr
cuddAllocNodePar(unique,index)
DdManager *unique;
int index;
{
    int i;
#ifdef __osf__
#pragma pointer_size save
#pragma pointer_size short
#endif
    DdNodeOld *page;
    DdNodeAdr Address;
    int pageNumber;
    DdMemory *memory;
    memory=&unique->memory[index];

    if (cuddEqualAdr(memory->nextFree,NULL_ADR)) 
      { /* memory is all full need to alloc a new page */
	if(memory->nextPage==NULL_PAGE)
	  return(NULL_ADR); /* No more pages available */ 

	if(memory->nextPage==memory->TableSize)
	  if(cuddResizePageTable(unique,index)==0)
	    return(NULL_ADR); /* unable to allocate mem for new entries */

	pageNumber=memory->nextPage;
	page = memory->PageTable[pageNumber] =ALLOC(DdNodeOld,DD_MEM_CHUNK)-1; 
	/* -1 to make it 1 indexed */ 
#ifdef __osf__
#pragma pointer_size restore
#endif
	if (page == NULL) {
#ifdef DD_VERBOSE
	  (void) fprintf(stderr,"cuddAllocNode: out of memory");
	  (void) fprintf(stderr,"Memory in use = %ld\n", unique->memused);
#endif
	  return(NULL_ADR);
	}

	if(unique->reordering)
	  {
	    DdNodeMore *ref_page;
	    ref_page=memory->PageTable2[pageNumber]=ALLOC(DdNodeMore,DD_MEM_CHUNK)-1;
	    if (ref_page == NULL) {
#ifdef DD_VERBOSE
	      (void) fprintf(stderr,"cuddAllocNode: out of memory");
	      (void) fprintf(stderr,"Memory in use = %ld\n", unique->memused);
#endif
	      return(NULL_ADR);
	    }
	    unique->memused += (DD_MEM_CHUNK) * sizeof(DdNodeMore);
	    for(i=1;i<=DD_MEM_CHUNK;i++)
	      ref_page[i].Reorder.Ref=0;
	  }
	    
	unique->memused += (DD_MEM_CHUNK) * sizeof(DdNodeOld);

	for(i=1;i<DD_MEM_CHUNK;i++)
	  {
	    page[i].next.page = pageNumber;
	    page[i].next.number = i+1;
	  } 

	page[DD_MEM_CHUNK].next = NULL_ADR;
	
	memory->nextFree.page = pageNumber;
	memory->nextFree.number=1; 

	memory->nextPage++;
      }

    unique->allocated++;
    Address = memory->nextFree;
    memory->nextFree = memory->PageTable[Address.page][Address.number].next;

    memory->allocated++;

    assert(Address.number!=0);
    return(Address);

} /* end of cuddAllocNode */

DdNodePar
cuddReorderAllocNodePar(DdManager *unique,int index)
{
    int i;
#ifdef __osf__
#pragma pointer_size save
#pragma pointer_size short
#endif
    DdNodeOld *page;
    DdNodeMore *ref_page;
    DdMemory *memory;
    int pageNumber;
    DdNodePar f;
    int mem_index;

    if (cuddEqualPar(unique->ReorderFreeList,CUDD_ERROR)) 
      { /* memory is all full */
	mem_index=cuddIndexToLevel(unique,index);
	/* we must choose some index where more space is to be allocated
	   the choise for now is the current level of the index od ne node */
	
	memory=&unique->memory[mem_index];
	
	if(memory->nextPage==NULL_PAGE)
	  return(CUDD_ERROR); 
	/* No more pages available (could use another index) */ 
	
	if(memory->nextPage==memory->TableSize)
	  if(cuddResizePageTable(unique,mem_index)==0)
	    return(CUDD_ERROR); /* unable to allocate mem for new entries */

	pageNumber=memory->nextPage;
	page = memory->PageTable[pageNumber] =ALLOC(DdNodeOld,DD_MEM_CHUNK)-1; 
	/* -1 to make it 1 indexed */ 
	ref_page
	  =memory->PageTable2[pageNumber]=ALLOC(DdNodeMore,DD_MEM_CHUNK)-1;

#ifdef __osf__
#pragma pointer_size restore
#endif
	if (page == NULL || ref_page==NULL) {
#ifdef DD_VERBOSE
	  (void) fprintf(stderr,"cuddAllocNode: out of memory");
	  (void) fprintf(stderr,"Memory in use = %ld\n", unique->memused);
#endif
	  return(CUDD_ERROR);
	}

	unique->memused += 
	  (DD_MEM_CHUNK) * (sizeof(DdNodeOld)+sizeof(DdNodeMore));

	for(i=1;i<DD_MEM_CHUNK;i++)
	  {
	    page[i].type.kids.T.index=mem_index;
	    page[i].type.kids.T.address.page = pageNumber;
	    page[i].type.kids.T.address.number = i+1;
	    ref_page[i].Reorder.Ref=0;
	  } 

	page[DD_MEM_CHUNK].type.kids.T = CUDD_ERROR;
	ref_page[DD_MEM_CHUNK].Reorder.Ref=0;
	
	unique->ReorderFreeList.address.page = pageNumber;
	unique->ReorderFreeList.address.number=1; 
	unique->ReorderFreeList.index=mem_index;
	memory->nextPage++;
      }
    f=unique->ReorderFreeList;
    unique->ReorderFreeList=Cudd_T_Par(unique,f);
    cuddReorderIndex(unique,f)=index;
    unique->allocated++;
    unique->memory[f.index].allocated++;

    assert(cuddLocal(unique,f.index));
    assert(f.address.number!=0);

    return(f);

} /* end of cuddReorderAllocNode */


void cuddDeallocNodePar(DdManager *manager,DdNodePar f)
{
  DdMemory *memory;
  memory=&manager->memory[f.index];
  if(manager->reordering)
    {
      Cudd_T_Par(manager,f)=manager->ReorderFreeList;
      manager->ReorderFreeList=f;
    }
  else
    {
      cuddNextPar(manager,f)=memory->nextFree;
      memory->nextFree=f.address;
    }      
  memory->allocated--;
  manager->allocated--;
} /* end of cuddDeallocNode */

void cuddDeleteNodePar(DdManager *manager,DdNodePar f)
{

  f.address.number=abs(f.address.number);
  /* both remove and dealloc only work with uncomplememted nodes */

  if(cuddEqualPar(Cudd_T_Par(manager,f),CUDD_FORWARD))
    {
      manager->ForwardNodes--;
    }
  else
    {
      cuddRemoveFromTable(manager,f); 
    }
  cuddDeallocNodePar(manager,f);

}

/* remove node from hash table */
static void cuddRemoveFromTable(DdManager *manager,DdNodePar f)
{
  int pos;
  DdNodeOld *looking_ptr,*f_ptr;
  DdNodePar T,E,looking;
  DdSubtable *subtable;
  
  subtable=&manager->subtables[cuddReorderIndex(manager,f)];
  f_ptr=cuddPtrPar(manager,f);
  T=f_ptr->type.kids.T;
  E=f_ptr->type.kids.E;
  
  pos = ddHashPar(T,E,subtable->slots);
  looking= subtable->ReorderNodelist[pos];

  if(cuddEqualPar(f,looking))
    {
      subtable->ReorderNodelist[pos].address=f_ptr->next;
      subtable->ReorderNodelist[pos].index
	=cuddReorderNextIndex(manager,f);
    }
  else
    {
      assert(!cuddEqualAdr(looking.address,NULL_ADR));
      looking_ptr=cuddPtrPar(manager,looking);
      while (!(cuddEqualAdr(f.address,looking_ptr->next) 
	       && cuddReorderNextIndex(manager,looking)==f.index))
	{
	  looking.index=cuddReorderNextIndex(manager,looking);
	  looking.address = looking_ptr->next;
	  assert(!cuddEqualAdr(looking.address,NULL_ADR));
	  looking_ptr=cuddPtrPar(manager,looking);
	}
      /* remove f */
      looking_ptr->next=f_ptr->next; 
      cuddReorderNextIndex(manager,looking)=cuddReorderNextIndex(manager,f);
    }

  subtable->keys--;
  manager->keys--;
}

void cuddAddToTable(DdManager *manager,DdNodePar f)
{
  int pos;
  DdNodeOld *f_ptr;
  DdNodePar T,E;

  f_ptr=cuddPtrPar(manager,f);
  T=f_ptr->type.kids.T;
  E=f_ptr->type.kids.E;
  
  pos = ddHashPar(T,E,manager->subtables[f.index].slots);
  f_ptr->next = manager->subtables[f.index].nodelist[pos];
  manager->subtables[f.index].nodelist[pos]=f.address;
  manager->subtables[f.index].keys++;
  manager->keys++;
}

void cuddClearSubtable(DdManager *manager,int index)
{
  DdSubtable *subtable;
  DdNodeAdr *nodelist;
  int j;

  subtable=&manager->subtables[index];
  nodelist = manager->subtables[index].nodelist;

  for (j = 0; (unsigned) j < subtable->slots; j++) {
    nodelist[j] = NULL_ADR;
  }
  subtable->keys=0;
}

int CuddInitIndex(DdManager *manager,int index)
{
  int j;
  DdNodeAdr *nodelist;
  DdNodePar *r_nodelist;

  manager->subtables[index].slots = CUDD_UNIQUE_SLOTS;
  manager->subtables[index].keys = 0;
  manager->subtables[index].maxKeys = CUDD_UNIQUE_SLOTS * DD_MAX_SUBTABLE_DENSITY;

  if(manager->reordering)
    {
      r_nodelist = manager->subtables[index].ReorderNodelist 
	= ALLOC(DdNodePar,CUDD_UNIQUE_SLOTS);
      if (r_nodelist == NULL) return(0);
      for (j = 0; (unsigned) j < CUDD_UNIQUE_SLOTS; j++) {
	r_nodelist[j] = CUDD_ERROR;
      }
    }
  else
    {
      nodelist = manager->subtables[index].nodelist 
	= ALLOC(DdNodeAdr,CUDD_UNIQUE_SLOTS);
      
      if (nodelist == NULL) return(0);
      for (j = 0; (unsigned) j < CUDD_UNIQUE_SLOTS; j++) {
	nodelist[j] = NULL_ADR;
      }
    }

  manager->memory[index].TableSize = DD_DEFAULT_PAGES;
  manager->memory[index].PageTable = ALLOC(DdNodeOld *,DD_DEFAULT_PAGES);
  if(manager->memory[index].PageTable == NULL)
    return(0);
  manager->memory[index].nextPage  = 0;
  manager->memory[index].nextFree  = NULL_ADR;
  manager->memory[index].allocated = 0;
  manager->allocatedPerLevel[index]= 0;
  
  manager->memused += sizeof(DdNodeAdr) * CUDD_UNIQUE_SLOTS +
    sizeof(DdNodeOld *) * DD_DEFAULT_PAGES;

  if(manager->reordering) {
    manager->memory[index].PageTable2 = ALLOC(DdNodeMore *,DD_DEFAULT_PAGES);
    manager->memused += sizeof(DdNodeMore *) * DD_DEFAULT_PAGES;
  }
  return(1);
}

/**Function********************************************************************

  Synopsis    [Creates and initializes the unique table.]

  Description [Returns a pointer to the table if successful.]

  SideEffects [None]

  Rev 3
  creates a uniquetable with DD_DEFAULT_RESIZE entries, each with 
  a default of CUDD_UNIQUE_SLOTS slots. 

  reason: simplicity over a minimal performance gain
  since the number of slots grows exponentially thre cost of having
  to rehash a few times is small

******************************************************************************/
DdManager *
cuddInitTablePar(int maxsize)
{
    DdManager	*unique = ALLOC(DdManager,1);
    int		i;
    unsigned int slots;

    unique->memused = sizeof(DdManager);
    unique->epsilon = DD_EPSILON;
    unique->maxGrowth = DD_MAX_REORDER_GROWTH;
    unique->size = 0;

    unique->maxSize = maxsize;


    /* Adjust the requested number of slots so as to decrease
    ** fragmentation. All table sizes will be drawn from a fixed
    ** sequence starting at CUDD_UNIQUE_SLOTS. This adjustment also
    ** guarantees that the number of slots will be prime.
    Rev3 changed
    */
    slots = CUDD_UNIQUE_SLOTS;

    unique->slots = 0;
    unique->keys = 0;

    unique->looseUpTo = DD_LOOSE_UP_TO;
    unique->gcEnabled = 1;  /* rev2 CHANGE to default no garbage collect */
    /* rev 3 on again */
    unique->allocated = 0;

    unique->subtables = ALLOC(DdSubtable,maxsize);
    if (unique->subtables == NULL) return(0);
    unique->memused += sizeof(DdSubtable)*maxsize;

    /* +++++++++++++++++++++++++++++ rev 3 memory managers */
    unique->memory = ALLOC(DdMemory,maxsize);
    if (unique->memory == NULL) return(0);
    unique->memused += sizeof(DdMemory)*maxsize;
       
    unique->allocatedPerLevel=ALLOC(int,maxsize);
    if (unique->allocatedPerLevel == NULL) return(0);
    unique->Local=ALLOC(int,maxsize);
    if (unique->Local == NULL) return(0);
    unique->perm = ALLOC(int,maxsize);
    if (unique->perm == NULL) return(0);
    unique->invperm = ALLOC(int,maxsize);
    if (unique->invperm == NULL) return(0);
    unique->OrigPerm = ALLOC(int,maxsize);
    if (unique->OrigPerm == NULL) return(0);
    unique->InvOrigPerm = ALLOC(int,maxsize);
    if (unique->InvOrigPerm == NULL) return(0);
    unique->memused += 5*sizeof(int)*maxsize;

    for(i =0; (unsigned) i < maxsize ; i++)
      {
	CuddInitIndex(unique,i);
	unique->perm[i] = i;
	unique->invperm[i] = i;
	unique->OrigPerm[i] = i;	
	unique->InvOrigPerm[i] = i;
      }
#ifdef __osf__
#pragma pointer_size save
#pragma pointer_size short
#endif
    unique->stack = ALLOC(DdNodePar ,(unique->maxSize)+1);
    unique->memused += sizeof(DdNodePar)*((unique->maxSize)+1);
#ifdef __osf__
#pragma pointer_size restore
#endif
    if (unique->stack == NULL) return(0);

    /* note the suntables and memory manages are not initialised */

    /* Initialize fields concerned with automatic dynamic reordering */
    unique->reorderings = 0;
    unique->autoDyn = 0;	/* initially disabled */

    unique->reordering = 0;
    unique->autoMethod = CUDD_REORDER_SIFT;

    unique->nextDyn = DD_FIRST_REORDER;
    unique->tree = NULL;

    unique->groupcheck = CUDD_GROUP_CHECK7;
    unique->recomb = DD_DEFAULT_RECOMB;
    unique->symmviolation = 0;
    unique->arcviolation = 0;
    unique->populationSize = 0;
    unique->numberXovers = 0;
    unique->linear = NULL;
    unique->linearSize = 0;

    unique->preGCHook = NULL;
    unique->postGCHook = NULL;
    unique->postReorderingHook = NULL;

    /* Initialize statistical counters. */
    unique->garbageCollections = 0;
    unique->totCachehits = 0;
    unique->cachecollisions = 0;
    unique->cacheinserts = 0;
    unique->cachedeletions = 0;

    return(unique);

} /* end of cuddInitTable */


/**Function********************************************************************

  Synopsis    [Frees the resources associated to a unique table.]

  Description []

  SideEffects [None]

******************************************************************************/
void
cuddFreeTablePar(unique)
DdManager *unique;
{
#ifdef __osf__
#pragma pointer_size save
#pragma pointer_size short
#endif
    int i;

    for(i=0;i<unique->size;i++)
      {
	while (unique->memory[i].nextPage != 0) 
	  FREE(unique->memory[i].PageTable[--unique->memory[i].nextPage]);
	FREE(unique->memory[i].PageTable);
	unique->memory[i].PageTable=NULL;
	unique->memory[i].TableSize=0;
      }
#ifdef __osf__
#pragma pointer_size restore
#endif
    
    for (i = 0; i < unique->size; i++) {
	FREE(unique->subtables[i].nodelist);
    }

    FREE(unique->subtables);

    FREE(unique->cache); 
    FREE(unique->perm);
    FREE(unique->OrigPerm);
    FREE(unique->InvOrigPerm);

    FREE(unique->vars);
    FREE(unique->stack);

/* REV2 nuked 
    if (unique->tree != NULL) Mtr_FreeTree(unique->tree);
*/

    if (unique->linear != NULL) FREE(unique->linear);
    FREE(unique);

} /* end of cuddFreeTable */

/**Function********************************************************************

  Synopsis    [Performs garbage collection on a unique table.]

  Description [Performs garbage collection on a unique table.
  If clearCache is 0, the cache is not cleared. This should only be
  specified if the cache has been cleared right before calling
  cuddGarbageCollect. (As in the case of dynamic reordering.)
  Returns the total number of deleted nodes.]

  SideEffects [None]

  SeeAlso     [cuddGarbageCollectZdd]

******************************************************************************/
int
cuddGarbageCollectPar(unique,clearCache)
DdManager *unique;
int clearCache;
{
  return(0);
}

DdNodePar cuddUniqueInterSafePar(unique,index,t,e)
DdManager *unique;
int index;
DdNodePar t;
DdNodePar e;
{
  DdNodePar res;
  if (cuddEqualPar(t,CUDD_ERROR) || (cuddEqualPar(e,CUDD_ERROR)))
    return(CUDD_ERROR);
  else
    {
      if (cuddEqualPar(t,e)) {
	res=t;
      } else {
	if (Cudd_IsComplementPar(t)) {
	  res = cuddUniqueInterPar(unique,index,Cudd_NotPar(t),Cudd_NotPar(e));
	  if (!cuddEqualPar(res,CUDD_ERROR))
	    res=Cudd_NotPar(res);
	} else {
	  res = cuddUniqueInterPar(unique,index,t,e);
	}
      }	
      return(res);
    }
}
/**Function********************************************************************

  Synopsis    [Checks the unique table for the existence of an internal node.]

  Description [Checks the unique table for the existence of an internal
  node. If it does not exist, it creates a new one.  Does not
  modify the reference count of whatever is returned.  A newly created
  internal node comes back with a reference count 0.  For a newly
  created node, increments the reference counts of what T and E point
  to.  Returns a pointer to the new node if successful; NULL if memory
  is exhausted or if reordering took place.]

  SideEffects [None]

  SeeAlso     [cuddUniqueInterZdd]

******************************************************************************/
DdNodePar
cuddUniqueInterPar(unique,index,T,E)
DdManager *unique;
int index;
DdNodePar T;
DdNodePar E;
{
    int pos;
#ifdef __osf__
#pragma pointer_size save
#pragma pointer_size short
#endif
    DdNodeAdr *nodelist;
    DdNodePar *r_nodelist;
    DdNodePar ret;
    DdNodeOld *looking_ptr,*node_ptr;
    DdSubtable *subtable;
#ifdef __osf__
#pragma pointer_size restore
#endif

#ifdef DD_UNIQUE_PROFILE
    cudd_UniqueLookUps++;
#endif

    subtable = &(unique->subtables[index]);

    if (subtable->keys > subtable->maxKeys) {
      ddRehash(unique,(int)index);
    }

    pos = ddHashPar(T,E,subtable->slots);
    if(unique->reordering)
      {
	r_nodelist = subtable->ReorderNodelist;
	ret = r_nodelist[pos];
	
	while (!cuddEqualAdr(ret.address,NULL_ADR)) {
	  looking_ptr=cuddPtrPar(unique,ret);
	  if (cuddEqualPar(looking_ptr->type.kids.T,T) 
	      && cuddEqualPar(looking_ptr->type.kids.E,E))
	    {
	      return(ret);  
	    }

	  ret.index=cuddReorderNextIndex(unique,ret);
	  ret.address = looking_ptr->next;
#ifdef DD_UNIQUE_PROFILE
	cudd_UniqueLinks++;
#endif
	}
	ret = cuddReorderAllocNodePar(unique,index);
	if (cuddEqualPar(ret,CUDD_ERROR)) return(CUDD_ERROR);
      	cuddIncreaseRefSafe(unique,T);
	cuddIncreaseRefSafe(unique,E);

	node_ptr=cuddPtrPar(unique,ret);
	node_ptr->next = r_nodelist[pos].address;
	cuddReorderNextIndex(unique,ret)=r_nodelist[pos].index;
	r_nodelist[pos] = ret;
      }
    else
      {
	ret.index=index;
	nodelist = subtable->nodelist;
	ret.address = nodelist[pos];
	
	while (!cuddEqualAdr(ret.address,NULL_ADR)) {
	  looking_ptr=cuddPtrPar(unique,ret);
	  if (cuddEqualPar(looking_ptr->type.kids.T,T) 
	      && cuddEqualPar(looking_ptr->type.kids.E,E))
	    {
	      return(ret);  
	    }
	  ret.address = looking_ptr->next;
#ifdef DD_UNIQUE_PROFILE
	cudd_UniqueLinks++;
#endif
	}
	ret.address = cuddAllocNodePar(unique,index);
	if (cuddEqualAdr(ret.address,NULL_ADR)) return(CUDD_ERROR);
	node_ptr=cuddPtrPar(unique,ret);
	node_ptr->next = nodelist[pos];
	nodelist[pos] = ret.address;
     }

    unique->keys++;
    subtable->keys++;

    node_ptr->type.kids.T = T;
    node_ptr->type.kids.E = E;
    
    return(ret);

} /* end of cuddUniqueInter */



/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Rehashes a unique subtable.]

  Description []

  SideEffects [None]

  SeeAlso     [ddRehashZdd]

******************************************************************************/
static void
ddRehash(unique,i)
DdManager *unique;
int i;
{
    unsigned int slots, oldslots;
    int j, pos;
#ifdef __osf__
#pragma pointer_size save
#pragma pointer_size short
#endif
    DdNodeAdr *nodelist, *oldnodelist;
    DdNodePar *r_nodelist, *r_oldnodelist;
#ifdef __osf__
#pragma pointer_size restore
#endif
    DdNodePar node;
    DdNodeOld *node_ptr;
    DdNodeAdr next;
    int next_index;

    /* Rev 3 NUKED gc must be determined otherwise 
    if (unique->reclaimed * 2 <= unique->allocated ||
    unique->slots > unique->looseUpTo) {
	unique->minDead = (DD_GC_PERCENT_LO * unique->slots) / 100;
#ifdef DD_VERBOSE
	if (unique->gcPercent == DD_GC_PERCENT_HI) {
	    (void) fprintf(stderr,"GC percentage = %d\t", DD_GC_PERCENT_LO);
	    (void) fprintf(stderr,"minDead = %d\n", unique->minDead);
	}
#endif
	unique->gcPercent = DD_GC_PERCENT_LO;
    }
    */

    assert(i != CUDD_MAXINDEX);
    oldslots = unique->subtables[i].slots;
    
    /* Compute the new size of the subtable. Normally, we just double
    ** and then make prime. However, after reordering, a table may be
    ** severely overloaded. We iterate so that all tables have sizes
    ** from a small set. This should minimize fragmentation.
    */
    slots = oldslots;
    do {
      slots = Cudd_Prime(DD_GROW_FACTOR * slots);
    } while (slots * DD_MAX_SUBTABLE_DENSITY < DD_GROW_FACTOR *
	     unique->subtables[i].keys);
    
    unique->subtables[i].slots = slots;
    unique->subtables[i].maxKeys = slots * DD_MAX_SUBTABLE_DENSITY;

    if(unique->reordering)
      {
	r_oldnodelist = unique->subtables[i].ReorderNodelist;
	r_nodelist = unique->subtables[i].ReorderNodelist 
	  = ALLOC(DdNodePar, slots);
	for (j = 0; (unsigned) j < slots; j++) {
	  r_nodelist[j] = CUDD_ERROR;
	}
	for (j = 0; (unsigned) j < oldslots; j++) {
	  node = r_oldnodelist[j];
	  while (!cuddEqualAdr(node.address, NULL_ADR)) {
	    node_ptr=cuddPtrPar(unique,node);
	    next = node_ptr->next;
	    next_index=cuddReorderNextIndex(unique,node);
	    pos = ddHashPar(node_ptr->type.kids.T,node_ptr->type.kids.E,
			    slots);
	    node_ptr->next = r_nodelist[pos].address;
	    cuddReorderNextIndex(unique,node)=r_nodelist[pos].index;
	    r_nodelist[pos] = node;
	    node.address = next;
	    node.index=next_index;
	  }
	}
	unique->memused += (slots - oldslots) * sizeof(DdNodePar);
	FREE(r_oldnodelist);
      }
    else
      {
	node.index=i;
	oldnodelist = unique->subtables[i].nodelist;
	nodelist = unique->subtables[i].nodelist = ALLOC(DdNodeAdr, slots);
	for (j = 0; (unsigned) j < slots; j++) {
	  nodelist[j] = NULL_ADR;
	}
	for (j = 0; (unsigned) j < oldslots; j++) {
	  node.address = oldnodelist[j];
	  while (!cuddEqualAdr(node.address, NULL_ADR)) {
	    node_ptr=cuddPtrPar(unique,node);
	    next = node_ptr->next;
	    pos = ddHashPar(node_ptr->type.kids.T,node_ptr->type.kids.E,
			    slots);
	    node_ptr->next = nodelist[pos];
	    nodelist[pos] = node.address;
	    node.address = next;
	  }
	}
	unique->memused += (slots - oldslots) * sizeof(DdNodeAdr);
	FREE(oldnodelist);
      }
    
#ifdef DD_VERBOSE
    (void) fprintf(stderr,"rehashing layer %d: keys %d dead %d new size %d\n",
		   i,unique->subtables[i].keys,unique->subtables[i].dead,slots);
#endif
    /* Update global data */
    unique->slots += (slots - oldslots);
    unique->maxCache = ddMin(unique->maxCacheHard,unique->slots); 
       
} /* end of ddRehash */

/**Function********************************************************************
  
  Synopsis [Increases the number of subtables in a unique table so
  that it meets or exceeds index.]

  Description [Increases the number of subtables in a unique table so
  that it meets or exceeds index. Returns 1 if successful; 0 otherwise.]

  SideEffects [None]

  SeeAlso     [cuddResizeTableZdd]

NOTE: Does not create variables, or initialize
******************************************************************************/
int ddResizeTablePar(unique,index)
DdManager *unique;
int index;
{
#ifdef __osf__
#pragma pointer_size save
#pragma pointer_size short
#endif
    DdSubtable *newsubtables;
    DdMemory *newmem;
#ifdef __osf__
#pragma pointer_size restore
#endif
    int oldsize,newsize;
    int i;
    int numSlots = CUDD_UNIQUE_SLOTS;
    int *newperm, *newinvperm, *neworigperm, *newinvorigperm,*newlocal,*newalloc;

    oldsize = unique->size;

    assert(index >= unique->maxSize); /* shoul not be called otherwise */
    /* The current table is too small: we need to allocate a new,
    ** larger one; move all old subtables, and initialize the new
    ** subtables up to index included.
    */
    newsize = index + DD_DEFAULT_RESIZE;
#ifdef DD_VERBOSE
    (void) fprintf(stderr,"Increasing the table size from %d to %d\n",
		   unique->maxSize, newsize);
#endif
    newsubtables = ALLOC(DdSubtable,newsize);
    if (newsubtables == NULL) return(0);
    newmem = ALLOC(DdMemory,newsize);
    if (newmem == NULL) return(0);
    newperm = ALLOC(int,newsize);
    if (newperm == NULL) return(0);
    newinvperm = ALLOC(int,newsize);
    if (newinvperm == NULL) return(0);
    neworigperm = ALLOC(int,newsize);
    if (neworigperm == NULL) return(0);
    newinvorigperm = ALLOC(int,newsize);
    if (newinvorigperm == NULL) return(0);
    newlocal = ALLOC(int,newsize);
    if (newlocal == NULL) return(0);
    newalloc = ALLOC(int,newsize);
    if (newalloc == NULL) return(0);

    /* NOTE needs update */
    unique->memused += 
      (newsize - unique->maxSize) * ((numSlots+1) * sizeof(DdNodeOld *) 
				     + 2 * sizeof(int) + sizeof(DdSubtable));
    
    /* copy old information */
    for (i = 0; i < oldsize; i++) {
      newsubtables[i]= unique->subtables[i];
      newmem[i]=unique->memory[i];
      newperm[i] = unique->perm[i];
      newinvperm[i] = unique->invperm[i];
      neworigperm[i] = unique->OrigPerm[i];
      newinvorigperm[i] = unique->InvOrigPerm[i];
      newlocal[i]=unique->Local[i];
      newalloc[i]=unique->allocatedPerLevel[i];
    }
    
    FREE(unique->subtables);
    unique->subtables = newsubtables;
    FREE(unique->memory);
    unique->memory = newmem;
    unique->maxSize = newsize;
    FREE(unique->perm);
    unique->perm = newperm;
    FREE(unique->invperm);
    unique->invperm = newinvperm;
    FREE(unique->OrigPerm);
    unique->OrigPerm = neworigperm;
    FREE(unique->InvOrigPerm);
    unique->InvOrigPerm = newinvorigperm;
    FREE(unique->Local);
    unique->Local=newlocal;
    FREE(unique->allocatedPerLevel);
    unique->allocatedPerLevel=newalloc;

    for (i = oldsize; i<newsize; i++)
      {
	CuddInitIndex(unique,i);
	unique->perm[i] = i;
	unique->invperm[i] = i;
	unique->OrigPerm[i] = i;	
	unique->InvOrigPerm[i] = i;
      }
	  
    unique->slots += (index + 1 - unique->size) * numSlots;
    unique->maxCache = ddMin(unique->maxCacheHard,unique->slots); 

    return(1);
    
} /* end of ddResizeTable */
    
int ddResizeVars(DdManager *unique)
{
    DdNode **newvars;
    int i;
#ifdef __osf__
#pragma pointer_size save
#pragma pointer_size short
#endif
    newvars = ALLOC(DdNode * ,unique->maxSize);
#ifdef __osf__
#pragma pointer_size restore
#endif
    if (newvars == NULL) return(0);
    for (i = 0; i < unique->size; i++) {
      newvars[i] = unique->vars[i];
    }

    FREE(unique->vars);
    unique->vars = newvars;
    return(1);
}
